"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Callable, Iterator, Tuple
from urllib.parse import urlencode


def _remove_none(obj: dict):
    keys = list(obj.keys())
    for k in keys:
        v = obj[k]
        if v is None:
            del obj[k]
    return obj


def with_query(url: str, **kwargs: Any) -> str:
    qs = urlencode(_remove_none(dict(kwargs)))
    if qs:
        if url.find("?") < 0:
            url += "?"
        else:
            url += "&"
        url += qs
    return url


class PageRequest:
    def __init__(self, fn_get_page: Callable, fn_filter: Callable = None, **kwargs):
        limit = kwargs.get("limit")
        self.limit = int(limit) if limit else 10
        self.fn_get_page = fn_get_page
        self.fn_filter = fn_filter
        self.query = _remove_none(dict(kwargs))
        if int(self.query.get("size", 0)) < 1:
            self.query["size"] = 20

    def get_page(self, page: int, size: int) -> list:
        content, _ = self.get_page(page, size)
        return content

    def _get_page(self, page: int, size: int) -> Tuple[list, bool]:
        params = dict(self.query)
        params["page"] = page
        params["size"] = size

        query_string = urlencode(params)
        page = self.fn_get_page(query_string)
        if not page or not page.content:
            return [], False
        if self.fn_filter:
            return list(filter(self.fn_filter, page.content)), True
        return page.content, True

    def get(self) -> list:
        ret = []
        page_index = self.query.get("page")
        if page_index is None:
            page_index = 0

        while True:
            self.query["page"] = page_index

            query_string = urlencode(self.query)
            page = self.fn_get_page(query_string)
            if not page or not page.content:
                break

            if self.fn_filter:
                content = list(filter(self.fn_filter, page.content))
            else:
                content = page.content
            ret += content
            if len(ret) > self.limit:
                ret = ret[: self.limit]
                break
            if len(page.content) < self.query["size"]:
                break  # no more items
            page_index += 1

        return ret

    def items(self) -> Iterator:
        page = 0
        size = self.query["size"]

        while True:
            data, has_next = self._get_page(page, size)
            if not has_next:
                break
            for item in data:
                yield item
            page += 1



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import pathlib
from typing import Any

from . import jsondot


class _StateFile:
    def __init__(self, file_path: str):
        self._path = file_path
        self._cache = None
        pathlib.Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)

    def _data(self, reload: bool = False):
        if self._cache is None or reload:
            self._cache = jsondot.load(self._path, {}, lock=True)
        return self._cache

    def get(self, key: str, default: Any, reload: bool = False):
        return self._data(reload).get(key, default)

    def set(self, key: str, value: Any):
        self._data()[key] = value
        jsondot.save(self._cache, self._path, lock=True)


_file: _StateFile = None


def init(store_path: str, name: str):
    global _file
    _file = _StateFile(os.path.join(store_path, name))


def get(key: str, default: Any, reload: bool = False):
    return _file.get(key=key, default=default, reload=reload)


def set(key: str, value: Any):
    _file.set(key, value)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

import click

from hcs_core.ctxp import CtxpException, recent
from hcs_core.ctxp.cli_options import *

org_id = click.option(
    "--org",
    type=str,
    default=None,
    required=False,
    help="Specify org ID. If not specified, default to the currently in-use org.",
)


def get_org_id(org: str = None) -> str:
    # 1st priority: user explicitly specified org
    with recent.of("org") as r:
        try:
            if org:
                return org

            # 2nd priority: environment variable
            org = os.environ.get("HCS_ORG")
            if org:
                return org

            # 3rd priority: last used
            org = r.get()
            if org:
                return org

            # 4th priority: org from the current auth token
            from hcs_core.sglib import auth

            auth_info = auth.details(False)
            if not auth_info:
                raise CtxpException("Not authorized. See 'hcs login --help'.")
            org = auth_info["org"]["id"]
            return org
        finally:
            if org:
                r.set(org)


def ensure_login():
    return get_org_id(None)



import logging
from weakref import WeakSet

log = logging.getLogger(__name__)


class Dispatcher:
    _registry: dict = {}

    def register(self, key: str, fn: callable):
        listeners = self._registry.get(key)
        if not listeners:
            listeners = WeakSet()
            self._registry[key] = listeners
        listeners.add(fn)
        return self

    def register_map(self, map: dict):
        for k, v in map.items():
            self.register(k, v)
        return self

    def unregister(self, key: str, fn: callable):
        listeners = self._registry.get(key)
        if listeners:
            listeners.remove(fn)
        return self

    def unregister_map(self, map: dict):
        for k, v in map.items():
            self.unregister(k, v)
        return self

    def emit(self, key: str, *args, **kwargs):
        listeners = self._registry.get(key)
        if listeners:
            copy = listeners.copy()
            for fn in copy:
                try:
                    fn(*args, **kwargs)
                except Exception as e:
                    log.error(f"Error handling event: {key}", e)
                    log.exception(e)
        return self

    def scope(self, events):
        return StrictDispatcher(self, events)


class StrictDispatcher:
    def __init__(self, impl: Dispatcher, events: set):
        self._impl = impl
        self._events = events

    def register(self, key: str, fn: callable):
        self._validate(key)
        self._impl.register(key, fn)
        return self

    def register_map(self, map: dict):
        for k, v in map.items():
            self.register(k, v)
        return self

    def unregister(self, key: str, fn: callable):
        self._validate(key)
        self._impl.unregister(key, fn)
        return self

    def unregister_map(self, map: dict):
        for k, v in map.items():
            self.unregister(k, v)
        return self

    def emit(self, key: str, *args, **kwargs):
        self._validate(key)
        self._impl.emit(key, *args, **kwargs)
        return self

    def _validate(self, key: str):
        if key not in self._events:
            raise Exception("Event is not in dispatcher scope: %s. The scoped events are: %s " % (key, self._events))



# Utility for Java Duration interoperability

import re

PATTERN = (
    "([-+]?)P(?:([-+]?[0-9]+)D)?(T(?:([-+]?[0-9]+)H)?(?:([-+]?[0-9]+)M)?(?:([-+]?[0-9]+)(?:[.,]([0-9]{0,9}))?S)?)?"
)

# Examples

# PT0.555S
# PT9M15S
# PT9H15M
# PT555H
# PT13320H


def to_seconds(duration_string: str):
    duration_string = duration_string.upper()

    p = re.compile(PATTERN)
    matcher = p.fullmatch(duration_string)
    if matcher is None:
        raise Exception("Unsupported duration format: %s" % duration_string)

    if matcher.start(3) >= 0 and matcher.end(3) == matcher.start(3) + 1 and duration_string[matcher.start(3)] == "T":
        raise Exception("Unsupported duration format: %s" % duration_string)

    day_start = matcher.start(2)
    day_end = matcher.end(2)
    hour_start = matcher.start(4)
    hour_end = matcher.end(4)
    minute_start = matcher.start(5)
    minute_end = matcher.end(5)
    second_start = matcher.start(6)
    second_end = matcher.end(6)

    total_seconds = 0
    if day_start >= 0:
        days = int(duration_string[day_start:day_end])
        total_seconds += days * 24 * 3600
    if hour_start >= 0:
        hours = int(duration_string[hour_start:hour_end])
        total_seconds += hours * 3600
    if minute_start >= 0:
        minutes = int(duration_string[minute_start:minute_end])
        total_seconds += minutes * 60
    if second_start >= 0:
        seconds = int(duration_string[second_start:second_end])
        total_seconds += seconds

    return total_seconds


if __name__ == "__main__":
    assert to_seconds("PT5S") == 5
    assert to_seconds("PT5M") == 5 * 60
    assert to_seconds("PT5H") == 5 * 3600
    assert to_seconds("PT2M3S") == 2 * 60 + 3
    assert to_seconds("PT2H3M") == 2 * 3600 + 3 * 60
    assert to_seconds("PT2H3S") == 2 * 3600 + 3
    assert to_seconds("PT2H3M4S") == 2 * 3600 + 3 * 60 + 4
    assert to_seconds("P1D") == 1 * 24 * 60 * 60
    assert to_seconds("P1DT1M") == 1 * 24 * 60 * 60 + 1 * 60
    assert to_seconds("P1DT1S") == 1 * 24 * 60 * 60 + 1
    assert to_seconds("P1DT1H1S") == 1 * 24 * 60 * 60 + 1 * 3600 + 1
    print("SUCCESS")



from . import context as context
from .actions import actions as actions
from .core import apply as apply
from .core import clear as clear
from .core import destroy as destroy
from .core import get_deployment_data as get_deployment_data
from .core import graph as graph
from .helper import PlanException as PlanException
from .helper import PluginException as PluginException
from .kop import attach_job_view as attach_job_view



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import re
from logging.handlers import RotatingFileHandler

import coloredlogs

LOG_FORMAT_SIMPLE = "%(levelname).4s %(asctime)s %(name)-16s %(message)s"
LOG_FORMAT_LONG = "%(color_on)s%(asctime)s.%(msecs)03d [%(process)-5d:%(threadName)-12s] %(levelname)-7s [%(name)-16s] %(message)s%(color_off)s"
DATE_FORMAT_SIMPLE = "%H:%M:%S"


def mask_password(msg: str) -> str:
    msg = re.sub(r"(\"password\"\s*:\s*\"?)(.*?)(\"?[\s*,? | \s*\}])", r"\1******\3", msg)
    return msg


class LogFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.CRITICAL: "\033[1;35m",
        logging.ERROR: "\033[1;31m",
        logging.WARNING: "\033[1;33m",
        logging.INFO: "\033[0;37m",
        logging.DEBUG: "\033[1;30m",
    }

    RESET_CODE = "\033[0m"

    def __init__(self, color, mask, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.color = color
        self.mask = mask

    def format(self, record, *args, **kwargs):
        if self.color is True and record.levelno in self.COLOR_CODES:
            record.color_on = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on = ""
            record.color_off = ""

        record.name = _shorten_package_name(record.name, 16)

        msg = super(LogFormatter, self).format(record, *args, **kwargs)
        return mask_password(msg) if self.mask else msg


class LogFormatterFixedNameWidth(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record, *args, **kwargs):
        record.name = _shorten_package_name(record.name, 16)
        return super(LogFormatterFixedNameWidth, self).format(record, *args, **kwargs)


def _shorten_package_name(package_name, max_length):
    if len(package_name) <= max_length:
        return package_name

    parts = package_name.split(".")
    shortened_parts = []

    for i, part in enumerate(parts):
        shortened_parts.append(part[0])

        if i == len(parts) - 1:
            # Last package name
            remaining_length = max_length - len(".".join(shortened_parts))
            if len(part) > remaining_length:
                shortened_parts[-1] = "..." + part[-remaining_length + 3 :]
            else:
                shortened_parts[-1] = part
        else:
            new_package_name = ".".join(shortened_parts)

            if len(new_package_name) + len(parts) - i - 1 > max_length:
                break

    return ".".join(shortened_parts)


def setup(
    console_log_output="stdout",
    console_log_level="INFO",
    console_log_color=True,
    console_log_mask=True,
    logfile_file=None,
    logfile_log_level="INFO",
    logfile_log_color=False,
    logfile_log_mask=True,
    log_line_template=LOG_FORMAT_SIMPLE,
    date_fmt=DATE_FORMAT_SIMPLE,
):
    # Alternative: 'global logger; logger = logging.getLogger("<name>")'
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    if console_log_output is not None:
        setup_console_output(
            logger,
            console_log_output,
            console_log_level,
            console_log_color,
            console_log_mask,
            log_line_template,
            date_fmt,
        )

    if logfile_file:
        setup_file_output(
            logger,
            logfile_file,
            logfile_log_level,
            logfile_log_color,
            logfile_log_mask,
            log_line_template,
        )


def setup_console_output(
    logger, console_log_output, console_log_level, console_log_color, console_log_mask, log_line_template, date_fmt
):
    if not date_fmt:
        date_fmt = coloredlogs.DEFAULT_DATE_FORMAT
    coloredlogs.install(
        level=console_log_level,
        fmt=log_line_template,
        datefmt=date_fmt,
        formatter=LogFormatter(fmt=log_line_template, color=console_log_color, mask=console_log_mask),
    )

    # console_log_output = console_log_output.lower()
    # if console_log_output == "stdout":
    #     console_log_output = sys.stdout
    # elif console_log_output == "stderr":
    #     console_log_output = sys.stderr
    # else:
    #     raise Exception("Failed to set console output: invalid output: '%s'" % console_log_output)
    # console_handler = logging.StreamHandler(console_log_output)

    # try:
    #     console_handler.setLevel(console_log_level.upper())
    # except:
    #     print("Failed to set console log level: invalid level: '%s'" % console_log_level)
    #     raise

    # console_formatter = LogFormatter(fmt=log_line_template, color=console_log_color, mask=console_log_mask)
    # console_handler.setFormatter(console_formatter)
    # logger.addHandler(console_handler)


def setup_file_output(
    logger,
    logfile_file,
    logfile_log_level,
    logfile_log_color,
    logfile_log_mask,
    log_line_template,
):
    logfile_max_bytes = 50 * 1024 * 1024
    logfile_max_count = 20
    try:
        logfile_handler = RotatingFileHandler(logfile_file, maxBytes=logfile_max_bytes, backupCount=logfile_max_count)
    except Exception as exception:
        print("Failed to set up log file: %s" % str(exception))
        raise

    try:
        logfile_handler.setLevel(logfile_log_level.upper())
    except Exception:
        print("Failed to set log file log level: invalid level: '%s'" % logfile_log_level)
        raise

    logfile_formatter = LogFormatter(fmt=log_line_template, color=logfile_log_color, mask=logfile_log_mask)
    logfile_handler.setFormatter(logfile_formatter)
    logger.addHandler(logfile_handler)


def get(file_name):
    return logging.getLogger(os.path.basename(os.path.splitext(file_name)[0]))



import re


def replace(text: str, mapping: dict, strict_on_unresolved_vars: bool = True, strict_on_unused_var: bool = False):
    unused_vars = set(mapping.keys())
    for k, v in mapping.items():
        new_text = text.replace("{{" + k + "}}", str(v))
        if new_text != text:
            text = new_text
            unused_vars.remove(k)

    if strict_on_unresolved_vars:
        unresolved_vars = re.findall(r"\{\{([^}]+)\}\}", text)
        if unresolved_vars:
            raise Exception(f"Strict mode: template variables unresolved: {unresolved_vars}")

    if strict_on_unused_var:
        if unused_vars:
            raise Exception(f"Strict mode: var specified but not in template: {unused_vars}")

    return text



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click

import hcs_core.ctxp as ctxp
import hcs_core.ctxp.cli_options as cli
import hcs_core.ctxp.util as util


@click.group(hidden=True)
def cli_test():
    """Commands for context. Each profile has its own context. The commands work for the current profile."""


@cli_test.command()
def echo_none():
    return


def _format_custom_table(data: list) -> str:
    fields_mapping = {"id": "Id", "opt": "Opt"}
    return util.format_table(data, fields_mapping)


@cli_test.command()
@click.option("--opt")
@click.argument("id")
def echo_obj(opt: str, id: str):
    return {"id": id, "opt": opt}


@cli_test.command()
@click.option("--opt")
@click.argument("obj_ids", nargs=-1)
@cli.formatter(_format_custom_table)
def echo_obj_list(opt: str, obj_ids: tuple[str]):
    ret = []
    for i in obj_ids:
        ret.append({"id": i, "opt": opt})
    return ret


@cli_test.command()
@click.argument("arg1", type=int)
def echo_int(arg1: int):
    return arg1


@cli_test.command()
@click.argument("arg1", type=int, nargs=-1)
def echo_int_list(arg1: tuple[int]):
    return list(arg1)


@cli_test.command()
@click.argument("arg1", type=str)
def echo_str(arg1: str):
    return arg1


@cli_test.command()
@click.argument("arg1", type=str, nargs=-1)
def echo_str_list(arg1: tuple[str]):
    return list(arg1)


@cli_test.command()
@click.argument("arg1", type=bool)
def echo_bool(arg1: bool):
    return arg1


@cli_test.command()
@click.argument("arg1", type=bool, nargs=-1)
def echo_bool_list(arg1: tuple[bool]):
    return list(arg1)


@cli_test.command()
@click.argument("arg1", type=float)
def echo_float(arg1: bool):
    return arg1


@cli_test.command()
@click.argument("arg1", type=float, nargs=-1)
def echo_float_list(arg1: tuple[bool]):
    return list(arg1)


@cli_test.command()
@click.option("--reason", type=str)
@click.option("--code", type=int)
@click.option("--ctxp-error/--tuple", type=bool)
def echo_error(reason: str, code: int, ctxp_error: bool):
    if ctxp_error:
        return ctxp.error(reason, code)
    return reason, code


@cli_test.command()
@click.option("--reason", type=str)
@click.option("--type", type=str)
def echo_exception(reason: str, type: str):
    root_cause = KeyError(reason)
    if type == "CtxpException":
        raise ctxp.CtxpException("pseudo-error") from root_cause
    if type == "TypeError":
        raise TypeError("pseudo-error") from root_cause
    if type == "httpx.HTTPStatusError":
        import httpx

        req = httpx.Request("GET", "http://ut")
        httpx.Response(400, text=reason, request=req).raise_for_status()
    if type == "Exception":
        raise Exception("pseudo-error") from root_cause
    raise Exception("Invalid exception type. This is an issue of the test case (caller).")



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
jsondot is utility to make json/dict object accessible in the "." way.

##########
#Example 1: load, update, and save JSON file
##########

data = jsondot.load('path/to/my.json')
print(data.hello.message)
data.hello.message = 'Hello, mortal.'
jsondot.save(data, 'path/to/my.json')


##########
#Example 2: decorate an existing python dict
##########

my_dict = jsondot.dotify(my_dict)
print(my_dict.key1.key2)
"""


import copy
import json
import os.path
import random
import time
from typing import Any, Dict, List, Optional, Union


class dotdict(dict):
    """dot.notation access to dictionary attributes with enhanced safety.

    A dictionary subclass that provides attribute-style access to its elements
    while maintaining compatibility with regular dict operations.

    Examples:
        >>> d = dotdict({'a': 1, 'b': {'c': 2}})
        >>> d.a
        1
        >>> d.b.c
        2
        >>> d.b.c = 3
        >>> d.b.c
        3
    """

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __lt__(self, other: Union[dict, "dotdict"]) -> bool:
        """Support proper sorting by comparing underlying dictionaries."""
        if isinstance(other, dict):
            return dict(self) < dict(other)
        return NotImplemented

    def __eq__(self, other: Union[dict, "dotdict"]) -> bool:
        """Support equality comparison with other dictionaries."""
        if isinstance(other, dict):
            return dict(self) == dict(other)
        return NotImplemented

    def __deepcopy__(self, memo: Optional[Dict] = None) -> "dotdict":
        """Support deep copying of the dictionary.

        Args:
            memo: Memoization dictionary for handling circular references

        Returns:
            A deep copy of the dotdict instance
        """
        new = {}
        for key in self.keys():
            new[key] = copy.deepcopy(self[key], memo=memo)
        return dotdict(new)

    def __repr__(self) -> str:
        """Return a string representation of the dictionary.

        Returns:
            A string showing this is a dotdict instance with its contents
        """
        return dict.__repr__(self)

    # def __getstate__(self):
    #     """Return state for serialization - just return the underlying dict"""
    #     return dict(self)

    # def __setstate__(self, state):
    #     """Restore state from serialization"""
    #     self.update(state)

    # def __reduce_ex__(self, protocol):
    #     return dict, (dict(self),)

    # def __getattribute__(self, name):
    #     """Make this look like a plain dict to introspection"""
    #     if name in ('__getstate__', '__setstate__', '__reduce__', '__reduce_ex__', '__class__'):
    #         return getattr(dict, name)
    #     return super().__getattribute__(name)


def _yaml_interoperability() -> None:
    """Set up YAML serialization support for dotdict objects.

    Registers a custom representer with PyYAML to properly serialize
    dotdict objects as regular dictionaries. This ensures dotdict
    objects can be seamlessly used with YAML operations.

    Note:
        This function is automatically called during module initialization.
    """
    try:
        import yaml

        def _represent_dict(dumper: yaml.SafeDumper, data: dotdict) -> yaml.nodes.MappingNode:
            return dumper.represent_dict(data.items())

        yaml.SafeDumper.add_representer(dotdict, _represent_dict)
    except ImportError:
        pass  # PyYAML not available, silently skip registration


def dotify(target: Any) -> Any:
    """Deeply convert an object from dict to dotdict.

    Recursively converts all nested dictionaries to dotdict instances
    while preserving other data types. Handles circular references safely.

    Args:
        target: The object to convert

    Returns:
        The converted object with all dicts replaced by dotdict instances

    Raises:
        RecursionError: If maximum recursion depth is exceeded
    """
    # Handle circular references with recursion depth check
    if getattr(dotify, "_depth", 0) > getattr(dotify, "_max_depth", 1000):
        raise RecursionError("Maximum recursion depth exceeded")

    dotify._depth = getattr(dotify, "_depth", 0) + 1
    try:
        if isinstance(target, dotdict):
            return target
        if isinstance(target, list):
            return [dotify(item) for item in target]
        if isinstance(target, dict):
            return dotdict({k: dotify(v) for k, v in target.items()})
        return target
    finally:
        dotify._depth -= 1


def undot(target: Any) -> Any:
    """Deeply convert an object from dotdict to plain dict.

    Recursively converts all nested dotdict instances to regular dictionaries
    while preserving other data types. Handles circular references safely.

    Args:
        target: The object to convert

    Returns:
        The converted object with all dotdict instances replaced by plain dicts

    Raises:
        RecursionError: If maximum recursion depth is exceeded
    """
    # Handle circular references with recursion depth check
    if getattr(undot, "_depth", 0) > getattr(undot, "_max_depth", 1000):
        raise RecursionError("Maximum recursion depth exceeded")

    undot._depth = getattr(undot, "_depth", 0) + 1
    try:
        if isinstance(target, list):
            return [undot(item) for item in target]
        if isinstance(target, dict):
            return {k: undot(v) for k, v in target.items()}
        return target
    finally:
        undot._depth -= 1


def _is_primitive(obj: Any) -> bool:
    """Check if the object is a primitive type.

    Determines if an object is a basic Python type that doesn't need
    conversion when serializing/deserializing.

    Args:
        obj: Any Python object

    Returns:
        bool: True if the object is a primitive type (str, bool, int, float, None)
    """
    return obj is None or isinstance(obj, (str, bool, int, float))


def plain(target: Any) -> Any:
    """Convert a complex object structure to plain Python types.

    Recursively converts all objects to their basic Python equivalents,
    making the structure suitable for serialization. Handles nested
    dictionaries, lists, and primitive types.

    Args:
        target: The object to convert

    Returns:
        The converted object with all complex types replaced by basic Python types

    Examples:
        >>> d = dotdict({'a': 1, 'b': {'c': 2}})
        >>> plain(d)
        {'a': 1, 'b': {'c': 2}}
    """
    if _is_primitive(target):
        return target
    if isinstance(target, list):
        return [plain(item) for item in target]
    if isinstance(target, dict):
        return {k: plain(v) for k, v in target.items()}
    return target


def parse(text: str) -> dotdict:
    """Parse a JSON string into a dotdict object.

    Args:
        text: JSON string to parse

    Returns:
        dotdict: The parsed and converted data

    Raises:
        JSONDecodeError: If the string contains invalid JSON
        TypeError: If input is not a string
    """
    if not isinstance(text, str):
        raise TypeError("Input must be a string")
    try:
        dict_data = json.loads(text)
        return dotify(dict_data)
    except json.JSONDecodeError as e:
        raise Exception(f"Invalid JSON: {str(e)}") from e


def _lock_with_retry(file: str, for_read: bool, max_retry: int = 6, initial_backoff: float = 0.05) -> None:
    """Attempt to lock a file with retries using exponential backoff.

    Args:
        file: File handle to lock
        for_read: If True, acquire shared lock; if False, acquire exclusive lock
        max_retry: Maximum number of retry attempts
        initial_backoff: Initial backoff time in seconds

    Raises:
        Exception: If unable to acquire lock after max retries
        portalocker.LockException: If locking fails
    """
    import portalocker

    flags = portalocker.LockFlags.SHARED if for_read else portalocker.LockFlags.EXCLUSIVE

    retry = 0
    backoff_seconds = initial_backoff
    while True:
        try:
            portalocker.lock(file, flags)
            return
        except portalocker.LockException as e:
            retry += 1
            if retry > max_retry:
                raise Exception(f"Failed to acquire lock for file {file} after {max_retry} attempts") from e
            random_delay = backoff_seconds + random.uniform(0, 0.1)
            time.sleep(random_delay)
            backoff_seconds *= 2


def load(file: str, default: Any = None, lock: bool = False) -> dotdict:
    """Load and parse a JSON file into a dotdict object.

    Args:
        file: Path to the JSON file
        default: Value to return if file doesn't exist
        lock: Whether to use file locking

    Returns:
        dotdict: The loaded and converted data

    Raises:
        FileNotFoundError: If file doesn't exist and no default provided
        JSONDecodeError: If file contains invalid JSON
        IOError: If file access fails
    """
    try:
        if not os.path.exists(file):
            if default is None:
                raise FileNotFoundError(f"File not found: {file}")
            return dotify(default)

        with open(file) as json_file:
            if lock:
                _lock_with_retry(json_file, True)
            dict = json.load(json_file)
        return dotify(dict)
    except json.JSONDecodeError as e:
        raise Exception(f"Invalid JSON in file {file}: {str(e)}") from e
    except IOError as e:
        raise Exception(f"Error accessing file {file}: {str(e)}") from e
    except Exception as e:
        raise Exception(f"Unexpected error loading file {file}: {str(e)}") from e


def save(data: dict, file: str, pretty: bool = True, lock: bool = False) -> None:
    """Save data as JSON to a file with optional pretty printing and locking.

    Args:
        data: Dictionary data to save
        file: Path to the output JSON file
        pretty: If True, format JSON with indentation
        lock: Whether to use file locking

    Raises:
        IOError: If file access fails
        TypeError: If data contains non-serializable objects
    """
    try:
        os.makedirs(os.path.dirname(file), exist_ok=True)
        with open(file, "w") as outfile:
            if lock:
                _lock_with_retry(outfile, False)
            json.dump(data, outfile, indent="\t" if pretty else None, default=vars)
    except IOError as e:
        raise Exception(f"Error writing to file {file}: {str(e)}") from e
    except TypeError as e:
        raise Exception(f"Error serializing data to JSON: {str(e)}") from e
    except Exception as e:
        raise Exception(f"Unexpected error saving file {file}: {str(e)}") from e



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click

import hcs_core.ctxp as ctxp
import hcs_core.ctxp.util as util


@click.group(cls=ctxp.cli_processor.LazyGroup)
def context():
    """Commands for context. Each profile has its own context. The commands work for the current profile."""


@context.command()
def list():
    """List all context item names."""
    return ctxp.context.list()


@context.command()
@click.argument("name")
@click.argument("key", required=False)
def get(name: str, key: str):
    """Get data of a specific context object."""
    data = ctxp.context.get(name)

    if key is None:
        return data
    if data is None:
        return None
    return data.get(key)


@context.command()
@click.argument("name")
@click.argument("key_value")  # 'key value pair, example: k1=v1'
def set(name: str, key_value: str):
    """Set a context object by name."""
    parts = key_value.split("=")
    if len(parts) != 2:
        ctxp.panic("Invalid KEY_VALUE format. Valid example: key1=value1")
    k, v = parts
    data = ctxp.context.get(name, default={})
    data[k] = v
    ctxp.context.set(name, data)


@context.command()
@click.argument("name")
def delete(name: str):
    """Delete a context object by name."""
    ctxp.context.delete(name)


@context.command()
def clear():
    """Delete all context objects."""
    ctxp.context.clear()


@context.command()
@click.argument("name", type=str, required=True)
def file(name: str):
    """Get the file path of the specific context object."""
    return _get_file(name)


def _get_file(name: str) -> str:
    return ctxp.context._store()._get_path(name)


@context.command()
@click.argument("name", type=str, required=True)
def edit(name: str):
    """Launch platform-specific editor to edit a context file."""
    util.launch_text_editor(_get_file(name))



import inspect
import os
import threading


def debounce(graceful_seconds):
    """
    Delay a function call by graceful_seconds. If an additional call happens during this time,
    the delay will be reset.
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            def call_impl():
                wrapper._timer = None
                return function(*args, **kwargs)

            # Cancel existing timer, if any
            if wrapper._timer is not None:
                wrapper._timer.cancel()

            # Schedule debounced run
            wrapper._timer = threading.Timer(graceful_seconds, call_impl)
            wrapper._timer.start()

        wrapper._timer = None
        return wrapper

    return decorator


def once(fn):
    """
    Raise exception if the function called more than once.
    """

    def wrapper(*args, **kwargs):
        if wrapper._invoked:
            raise Exception("Duplicated invocation: %s, %s" % (_get_caller_file(), fn.__name__))
        wrapper._invoked = True
        return fn(*args, **kwargs)

    wrapper._invoked = False
    return wrapper


def _get_caller_file():
    # first get the full filename (including path and file extension)
    caller_frame = inspect.stack()[2]
    caller_filename_full = caller_frame.filename

    # now get rid of the directory (via basename)
    # then split filename and extension (via splitext)
    caller_filename_only = os.path.splitext(os.path.basename(caller_filename_full))[0]

    # return both filename versions as tuple
    return caller_filename_only



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Profile_store provides a utility method to create profile-scoped fstore

import os

from . import profile
from .fstore import fstore
from .util import CtxpException

_store_map: dict[str, fstore] = {}
_active_profile_name: str = None


def profile_store(store_name: str) -> fstore:
    global _store_map, _active_profile_name

    profile_name = profile.name()
    if not profile_name:
        raise CtxpException("Profile not specified")

    if profile_name != _active_profile_name:
        # profile changed
        _active_profile_name = profile_name
        _store_map = {}

    ret = _store_map.get(store_name)
    if ret is None:
        ret = _store_from_profile_name(profile_name, store_name)
        _store_map[store_name] = ret
    return ret


def _store_from_profile_name(profile_name: str, store_name: str) -> fstore:
    profile_path = profile.path(profile_name)
    store_path = os.path.join(profile_path, store_name)
    return fstore(store_path)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import hashlib
import json
import logging
import time

import jwt

from hcs_core.ctxp import CtxpException, panic, profile
from hcs_core.ctxp.jsondot import dotdict, dotify

from .csp import CspClient
from .login_support import create_oauth_client, refresh_oauth_token

log = logging.getLogger(__name__)


def _get_profile_auth_hash(effective_profile):
    csp = effective_profile.csp
    text = json.dumps(csp, default=vars)
    return profile.name() + "#" + hashlib.md5(text.encode("ascii"), usedforsecurity=False).hexdigest()


def _is_auth_valid(auth_data, effective_profile):
    if not auth_data:
        return
    if not auth_data.token:
        return
    if auth_data.hash != _get_profile_auth_hash(effective_profile):
        return
    if time.time() > auth_data.token.expires_at:
        return
    return True


def _decode_http_basic_auth_token(basic_token: str):
    import base64

    try:
        decoded = base64.b64decode(basic_token).decode("utf-8")
        client_id, client_secret = decoded.split(":")
        return client_id, client_secret
    except Exception as e:
        raise CtxpException(f"Invalid basic http auth token: {e}")


def login(force_refresh: bool = False, effective_profile=None):
    """Ensure login state, using credentials from the current profile. Return oauth token."""
    # _validate_profile_readiness()

    if effective_profile is None:
        effective_profile = profile.current()

    auth_data = profile.auth.get()  # TODO
    if force_refresh or not _is_auth_valid(auth_data, effective_profile):
        oauth_token = _get_new_oauth_token(auth_data.token, effective_profile)
        if oauth_token:
            use_oauth_token(oauth_token, effective_profile)
    else:
        oauth_token = auth_data.token
    return oauth_token


def _get_new_oauth_token(old_oauth_token, effective_profile):
    csp_config = effective_profile.csp

    csp_client = CspClient(url=csp_config.url)

    if csp_config.apiToken:
        oauth_token = csp_client.login_with_api_token(csp_config.apiToken)
    elif csp_config.clientId:
        oauth_token = csp_client.login_with_client_id_and_secret(
            csp_config.clientId, csp_config.clientSecret, csp_config.orgId
        )
    elif csp_config.basic:
        client_id, client_secret = _decode_http_basic_auth_token(csp_config.basic)
        oauth_token = csp_client.login_with_client_id_and_secret(client_id, client_secret, csp_config.orgId)
    else:
        # This should be a config from interactive login.
        # Use existing oauth_token to refresh.
        if not old_oauth_token:
            old_oauth_token = profile.auth.get().token
        if old_oauth_token:
            try:
                oauth_token = refresh_oauth_token(old_oauth_token, csp_config.url)
            except Exception as e:
                oauth_token = None
                log.warning(e)
        else:
            oauth_token = None
    return oauth_token


def oauth_client():
    return _oauth_client_impl(profile.current())


def _oauth_client_impl(effective_profile):
    oauth_token = login(False, effective_profile=effective_profile)

    if not oauth_token:
        panic(
            "Login failed. If this is configured API key or client credential, refresh the credential from CSP and update profile config. If this is browser based interactive login, login again."
        )

    csp_url = effective_profile.csp.url

    def fn_on_new_oauth_token(token, refresh_token=None, access_token=None):
        use_oauth_token(token, effective_profile)

    def fn_get_new_oauth_token():
        return _get_new_oauth_token(old_oauth_token=None, effective_profile=effective_profile)

    if effective_profile.csp.apiToken:
        # for api-token login, due to no client ID, it will fail with 400 during refresh.
        # So instead of doing standard OAuth refresh, just use the API token to acquire a new one.
        return create_oauth_client(oauth_token, csp_url, fn_on_new_oauth_token, fn_get_new_oauth_token)
    else:
        # return create_oauth_client(oauth_token, csp_url, fn_on_new_oauth_token)

        # With the updated CSP (Omnissa), the default refresh mechanism from oauth client does not work, it might be the service
        # side issue.
        # To make it compatible, use the custom refresh mechanism (get new token).
        return create_oauth_client(oauth_token, csp_url, fn_on_new_oauth_token, fn_get_new_oauth_token)


def details(get_org_details: bool = False) -> dotdict:
    """Get the auth details, for the current profile"""
    oauth_token = login()
    if not oauth_token:
        return
    return details_from_token(oauth_token, get_org_details)


def details_from_token(oauth_token, get_org_details: bool = False):
    decoded = jwt.decode(oauth_token["access_token"], options={"verify_signature": False})
    org_id = decoded["context_name"]
    ret = {"token": oauth_token, "jwt": decoded, "org": {"id": org_id}}

    if get_org_details:
        csp_client = CspClient(url=profile.current().csp.url, oauth_token=oauth_token)
        try:
            org_details = csp_client.get_org_details(org_id)
        except Exception as e:
            org_details = {"error": f"Fail retrieving org details: {e}"}
        ret["org"].update(org_details)
    return dotify(ret)


def get_org_id_from_token(oauth_token: str) -> str:
    decoded = jwt.decode(oauth_token["access_token"], options={"verify_signature": False})
    return decoded["context_name"]


def use_oauth_token(oauth_token, effective_profile=None):
    if effective_profile is None:
        effective_profile = profile.current()

    if "expires_at" not in oauth_token:
        safe_expires_in = (
            20 * 60  # use 20 minutes which is
        )  # Workaround current CSP inconsistency, which reports the expiry as 4 hrs but actually expires in 30 mins.
        expires_in = min(oauth_token["expires_in"], safe_expires_in)
        oauth_token["expires_at"] = int(time.time()) + expires_in
    else:
        # reduce the expiry time by 10 minutes.
        told = int(oauth_token["expires_at"])
        adjusted = told - 10 * 60
        adjusted = max(adjusted, time.time() + 60)  # at least 1 minute in future to avoid frenzy refresh.
        oauth_token["expires_at"] = adjusted
    profile.auth.set({"token": oauth_token, "hash": _get_profile_auth_hash(effective_profile)})
